
#include <list>
#include <string.h>
#include <stdlib.h>
#include <setjmp.h>
#include <stdio.h>
#include <cassert>

#include "scheduler.h"
#include "mangling.h"

using namespace std;

typedef struct {
	int value;
	int pid;
	jmp_buf env;            //environment for setjmp/longjmp
	int (*func)(void *);    //pointer to thread function
	void *arg;              //arguments = local data
	int running;            //current state: start, running, finished
} Process_t;

#define TABLE_SIZE ((MAX_PROCESSES-1) + 1 + MAX_VALUE)
#define IDX_TO_VALUE(idx) (idx - (MAX_PROCESSES-1))
#define VALUE_TO_IDX(value) (value + (MAX_PROCESSES-1))

static list<Process_t*> sched_table[TABLE_SIZE];

static Process_t *current = 0;

static jmp_buf sched_env;

static int next_free_pid = 0;

static inline int sched_random() {
	return rand() % (MAX_VALUE+1);
	//return 0;
}

void sched_yield(void) {
    if ( !setjmp(current->env) ) 
            longjmp(sched_env,1);
}


int sched_init(void) {
	return 0;
}

static void print_content(void) {
	int idx = 0;
	for (idx = TABLE_SIZE-1; idx >= 0; idx--) {
		int value = IDX_TO_VALUE(idx);
		printf("%4d:",value);
		for (list<Process_t*>::iterator iter = sched_table[idx].begin();
		     iter != sched_table[idx].end();
		     iter++) {
			if (iter != sched_table[idx].begin()) {
				printf(",");
			}
			printf("%d",(*iter)->pid);
		}
		printf("\n");
	}
}

void schedule(void) {
	do {
		int table_idx = 0;
		current = 0;
		
		print_content();
		for (table_idx = 0; table_idx < TABLE_SIZE && !current; ++table_idx) {
			int value = IDX_TO_VALUE(table_idx);
			size_t size = sched_table[table_idx].size();
			if (size) {
				int sel = rand() % size;
				int i = 0;
				list<Process_t*>::iterator iter;

				for (i = 0, iter = sched_table[table_idx].begin();
				     i < sel;
				     i++, iter++);
				current = *iter;
				assert(current->value == value);
				sched_table[table_idx].erase(iter);
				
				printf("selected process %d with value %d\n", current->pid, current->value);

				for (i = table_idx; i < TABLE_SIZE; ++i ) {
					while (sched_table[i].size()) {
					       Process_t *p = sched_table[i].front();
					       sched_table[i].pop_front();
					       p->value--;
					       sched_table[i-1].push_back(p);
					       assert(p->value == IDX_TO_VALUE(i-1));
					}
				}

				if (!setjmp(sched_env)) 
					longjmp(current->env,1);

				if (current->running) {
					current->value = sched_random();
					sched_table[VALUE_TO_IDX(current->value)].push_back(current);
				}
			}
		}

	} while (current);
}


int sched_pid(void) {
	if (!current)
		return -1;
	return current->pid;
}


static inline void sched_wrapper(void) {
	// start function
	current->func(current->arg);

	// if the function has terminated, we are at this point 
	// set flag that indicates that function has terminated
	current->running = 0;

	// function has finished, continue with the scheduler
	sched_yield();
}


int sched_register(int (*func)(void *), 
		   int sk_size,
		   void *arg) {
	Process_t *p;
	void *stackPtr;

	p = (Process_t*) malloc(sizeof(Process_t));
	if (!p) {
		return -1;
	}
	stackPtr = calloc(1,sk_size);
	if (!stackPtr) {
		free(p);
		return -1;
	}
	memset(p, 0, sizeof(Process_t));

	p->pid = next_free_pid++;
	p->value = sched_random();
	p->func = func;
	p->arg = arg;
	p->running = 1;
	
	sched_table[VALUE_TO_IDX(p->value)].push_back(p);
	
	// the following is now prepared (dependent on the environment)
	// 1) the instruction pointer is set to the wrapper function, here
	//    starts and ends our thread
	// 2) the stack base pointer is set to the beginning of our stack
	// 3) the stack pointer is set to the top of the stack
	// 4) the top of the stack is the old instruction pointer
	//    when a function returns, the program is going back to this address
	//    as the wrapper handles the termination of our threads, we set
	//    this to 0

#if defined(__linux__) && defined(__i386__)
	p->env[0].__jmpbuf[5] = (int)sched_wrapper; /* EIP */
	PTR_MANGLE(p->env[0].__jmpbuf[5]);
	p->env[0].__jmpbuf[3] = 0; /* EBP */
	PTR_MANGLE(p->env[0].__jmpbuf[3]);
	p->env[0].__jmpbuf[4] = (int)stackPtr + sk_size - sizeof(int32_t); /* ESP */
	*((void**)p->env[0].__jmpbuf[4]) = (void *)0; /* old EIP */
	PTR_MANGLE(p->env[0].__jmpbuf[4]);
#elif defined(__linux__) && defined(__x86_64__)
	p->env[0].__jmpbuf[7] = (long int)sched_wrapper; /* RIP */
	PTR_MANGLE(p->env[0].__jmpbuf[7]);
	p->env[0].__jmpbuf[1] = 0; /* RBP */
	PTR_MANGLE(p->env[0].__jmpbuf[1]);
	p->env[0].__jmpbuf[6] = (long int)stackPtr + sk_size - sizeof(int64_t); /* RSP */
	*((void**)p->env[0].__jmpbuf[6]) = (void *)0; /* old RIP */
	PTR_MANGLE(p->env[0].__jmpbuf[6]);
#elif defined(__linux__) && defined(__PPC__)
	p->env[0].__jmpbuf[2] = (int)sched_wrapper; /* LR */
	p->env[0].__jmpbuf[0] = (int)stackPtr + sk_size - 32 * sizeof(int); /* SP */
	*((int *)p->env[0].__jmpbuf[0]) = 0; /* old SP */
	*((int *)p->env[0].__jmpbuf[0] + sizeof(int)) = 0; /* old LR */
	PTR_MANGLE(p->env[0].__jmpbuf[0]);
	PTR_MANGLE(p->env[0].__jmpbuf[2]);
#elif defined(__MACH__) && defined(__ppc__)   /* ppc32-osx */
	p->env[0] = ((int)stackPtr + sk_size - 32 * sizeof(int)); /* SP */
	*((int *)p->env[0]) = 0; /* old SP */
	*((int *)p->env[0] + sizeof(int)) = 0; /* old LR */
	p->env[21] = (int)sched_wrapper; /* LR */
#elif defined(__MACH__) && defined(__i386__)   /* i386-osx */
	p->env[12] = (int)sched_wrapper; /* EIP */
	p->env[8] = 0; /* EBP */
	p->env[9] = (int)stackPtr + sk_size - sizeof(int32_t); /* ESP */
	*((void**)p->env[9]) = (void *)0; /* old EIP */
#endif 
	return 0;
}
